from ceda_flight_pipeline.flight_client import resolve_link
import json

def update_moles(client, thorough=False, cache=True):
    # Get list of all ids
    # Find catalogue link for each
    # Add catalogue link or replace value in existing field.

    push_records = []

    stable, updated_link = 0, 0

    attempted = {}

    records = client.obtain_ids()

    for x, r in enumerate(records):
        new_link = None
        if thorough or 'catalogue_link' not in r['_source']:
            dpath = r['_source']['description_path']
            if dpath not in attempted:
                print(dpath)
                new_link = resolve_link(r['_source']['description_path'])
                attempted[dpath] = new_link
            else:
                new_link = attempted[dpath]

        if new_link:
            updated_link += 1
            r['_source']['catalogue_link'] = f'https://catalogue.ceda.ac.uk/uuid/{new_link}'
            push_records.append(r['_source'])
            # Cache new record
            if cache:
                with open(f'localcache/{x}.json','w') as f:
                    f.write(json.dumps(r))
        else:
            stable += 1

    print(f'Links updated: {updated_link}, Stable: {stable}')
    if updated_link > 0:
        proceed = input('Proceed to update? (y/n): ')
        if proceed == 'y':
            client.bulk_push(push_records)
    else:
        print('No links updated')
    

        


import unittest
import os
from unittest.mock import patch, mock_open

# Importing the functions to be tested from cli.py
from ceda_flight_pipeline.cli import str2bool


class TestFlightUpdate(unittest.TestCase):

    # Test the conversion function, turnings strings into bools

    def test_str2bool(self):

        self.assertTrue(str2bool("yes"))
        self.assertTrue(str2bool("Y"))
        self.assertTrue(str2bool("1"))
        self.assertFalse(str2bool("no"))
        self.assertFalse(str2bool("false"))

    # Test openConfig() when the config file isn't set up correctly and there aren't enough lines -> triggering an index error
    @patch(
        "builtins.open", new_callable=mock_open, read_data="line0\n/flights-dir\n"
    )  # Only two lines
    @patch.dict(os.environ, {"CONFIG_FILE": "dummy_config"})
    def test_open_config_index_error(self, mock_file):
        from ceda_flight_pipeline.cli import openConfig

        with patch("builtins.print") as mock_print:
            root, archive = openConfig()
        # Check if it returns empty strings as expected
        self.assertEqual((root, archive), ("", ""))

        mock_print.assert_called_once()

    # Test moveOldFiles() when files have to be moved
    def test_move_old_files_move_mode(self):
        commands = []  # initialising empty list where future os commands will be added

        # Capture shell commands by os.system
        def check_cmnds(cmd):
            commands.append(cmd)

        # Creating test files and dirs
        files = ["test.json", "/nested/file2.json"]
        root = "/flights-dir"
        archive = "/archive"

        with patch("os.system", check_cmnds):
            from ceda_flight_pipeline.cli import moveOldFiles

            moveOldFiles(root, archive, files)

        # Expected operations for each test file from move function
        expected = [
            f"mv /flights-dir/test.json /archive/test.json",
            f"mv /flights-dir/file2.json /archive/file2.json",
        ]
        # Check the commands are as expected
        self.assertEqual(commands, expected)

    # Test moveOldFiles() when archive mode is DELETE (files should be deleted)
    def test_move_old_files_delete_mode(self):
        commands = []  # initialising empty list where future os commands will be added

        # Capture shell commands by os.system
        def check_cmnds(cmd):
            commands.append(cmd)

        # Creating test files and dirs
        files = ["test.json", "/nested/file2.json"]
        root = "/flights-dir"
        archive = "DELETE"

        with patch("os.system", check_cmnds):
            from ceda_flight_pipeline.cli import moveOldFiles

            moveOldFiles(root, archive, files)

        # Expected delete operations for each file

        expected = [f"rm /flights-dir/test.json", f"rm /flights-dir/file2.json"]

        self.assertEqual(commands, expected)  # check correct


# test runner
if __name__ == "__main__":
    unittest.main()



__author__ = "Ioana Circu"
__contact__ = "ioana.circu@stfc.ac.uk"
__copyright__ = "Copyright 2025 United Kingdom Research and Innovation"


import click
import sys

from ceda_flight_pipeline.logger import setup_from_config
import logging
import os
import glob
from ceda_flight_pipeline.flight_client import ESFlightClient

logger = logging.getLogger(__name__)
from ceda_flight_pipeline.utils import logstream
logger.addHandler(logstream)
logger.propagate = False

IS_FORCE = True
VERB = True

# Helper function for converting string to boolean
def str2bool(v):
    """
    Input parameter: Str
    Returns: Bool based on whether the string is part of list
    """
    return v.lower() in ("y", "yes", "true", "t", "1")

@click.group()
def main():
    """Command Line Interface for flight update"""
    pass


@main.command()
@click.option(
    "--archive_path",
    default=None,
    help="Set path for archiving pushed flights",
)

@click.option(
    "--new_flights_dir",
    default=None,
    required=False,
    help="Set path to new set of flights",
)

@click.option(
    "--update",
    default=None,
    type=str,
    help="Name of script in updates/ to use",
)

@click.option(
    "--reindex",
    default=None,
    type=str,
    help="New elasticsearch index to move to",
)

@click.option(
    "--refresh",
    "-r",
    default=0,
    count=True,
    type=int,
    help="Refresh records, namely to identify catalogue links. Add extra `-r` flags to be more thorough",
)

@click.option(
    "--config_file",
    default=None,
    type=str,
    help="Path to config file",
)

@click.option(
    "--settings_file",
    default=None,
    type=str,
    help="Path to settings file",
)

@click.option(
    "--stac_template",
    default=None,
    type=str,
    help="Path to STAC template",
)

@click.option(
    "--verbose",
    "-v",
    default=0,
    count=True,
    type=int,
    help="Verbosity",
)

@click.option(
    "--console_log",
    default=True,
    type=bool,
    help="Log to console",
)

@click.option(
    "--stac_index",
    default=None,
    type=str,
    help="Main STAC index to reference",
)

@click.option(
    "--keep_files",
    is_flag=True,
    default=False,
    help="Move files after pushing",
)

def flight_update(
        archive_path, 
        new_flights_dir, 
        update, 
        reindex,
        config_file, 
        settings_file, 
        stac_template,
        verbose,
        console_log,
        stac_index,
        keep_files,
        refresh):
    """
    Main function running the flight update scripts based on the given command line parameters
    """

    move_files = not keep_files

    REPUSH = False
    add = True

    if update is not None or reindex is not None:
        add = False

    if not config_file:
        config_file = os.environ.get("FLIGHT_CONFIG")

    if not settings_file:
        settings_file = os.environ.get("FLIGHT_CONNECTION")
    
    if not stac_template:
        stac_template = os.environ.get("STAC_TEMPLATE")

    new_flights_dir, archive_path, stac_index = setup_from_config(
        new_flights=new_flights_dir,
        archive=archive_path,
        verbose=verbose,
        console_logging=console_log,
        stac_index=stac_index,
        cfg_file=config_file)
    
    logger.info("Setting up Flight Client interface")
    
    fclient = ESFlightClient(stac_index, settings_file, stac_template=stac_template)

    if refresh > 0:
        logger.info("Performing index refresh for elasticsearch")
        fclient.refresh(refresh)
        return

    if add:
        # Ensure archive_path and new_flights_dir are not empty
        if not archive_path:
            logger.error("Please provide an archive path.")
            return
        elif not new_flights_dir:
            logger.error("Please provide a directory for flights.")
            return
        else:
            fclient.addFlights(
                archive_path,
                new_flights_dir, 
                repush=REPUSH,
                move_files=move_files)

    elif update is not None:
        fclient.updateFlights(update)

    elif reindex is not None:
        fclient.reindex(reindex)

    else:
        logger.error("Mode unrecognized. Please choose either add or update.")
        return


if __name__ == "__main__":
    main()



__author__    = "Ioana Circu"
__contact__   = "ioana.circu@stfc.ac.uk"
__copyright__ = "Copyright 2025 United Kingdom Research and Innovation"


import logging
import os

logger = logging.getLogger(__name__)
from ceda_flight_pipeline.utils import logstream, formatter
logger.addHandler(logstream)
logger.propagate = False

levels = [
    logging.WARN,
    logging.INFO,
    logging.DEBUG,
]


def setup_logging(
        verbose: int = 0, 
        console_logging: bool = True, 
        log_file: str = "") -> None:
    """
    Sets up logging configuration. If `enable_logging` is False, no logging will occur.
    
    :param enable_logging: Flag to enable/disable logging.
    """
    fh = None
    log_file = log_file or ''

    if log_file != '':
        fh = logging.FileHandler(log_file),  # Write output to file
        fh.setLevel(levels[verbose])
        fh.setFormatter(formatter)

    for name in logging.root.manager.loggerDict:
        lg = logging.getLogger(name)
        lg.setLevel(levels[verbose])
        if fh is not None:
            lg.addHandler(fh)
        lg.debug(f'Reset handler for {name}')

def get_config(cfg_file) -> tuple:
    """
    Function to get logging info from config file
    """

    cfg_file = cfg_file or ''

    try:
        with open(cfg_file) as f: # 'r' is default if not specified.
            content = [r.strip() for r in f.readlines() if not r.startswith('#')] # Removes the '\n' from all lines
    
    except FileNotFoundError:
        logger.info("Config file not found.")
        return None, None, None, None, None, None
    
    new_flights = content[0]
    archive     = content[1]
    logfile     = content[2]

    values = {i.split('=')[0] : i.split('=')[1] for i in content[3:]}

    verbose = values.get('verbose',None)
    console_logging = values.get('console_logger',None)
    stac_index = values.get('index',None)

    return new_flights, archive, logfile, verbose, console_logging, stac_index

def setup_from_config(
        new_flights: str = None, 
        archive: str = None, 
        logfile: str = None, 
        verbose: int = None, 
        console_logging: bool = None, 
        stac_index: str = None,
        cfg_file: str = 'dirconfig'):

    new_flights_d, archive_d, logfile_d, verbose_d, console_logging_d, stac_index_d = get_config(cfg_file)

    # Apply defaults from config
    new_flights = new_flights or new_flights_d
    archive = archive or archive_d
    logfile = logfile or logfile_d
    verbose = verbose or verbose_d
    console_logging = console_logging or console_logging_d
    stac_index = stac_index or stac_index_d

    # Set up logging with a flag (True to enable logging, False to disable logging)
    setup_logging(verbose, console_logging, logfile)  # Change to False to disable logging

    return new_flights, archive, stac_index




'''
 --- Elastic Client ---
  - Extract all items from flight finder index by organisation
  - Save flight numbers to array - return array
'''

import json
import os, sys
import numpy as np
import glob

import logging

from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

from ceda_flight_pipeline.logger import logger

from ceda_flight_pipeline.simple_client import SimpleClient, gen_id

from datetime import datetime

import re
import requests

import urllib3
urllib3.disable_warnings()

logger = logging.getLogger(__name__)
from ceda_flight_pipeline.utils import (
    logstream, 
    STAC_TEMPLATE, 
    SPECIFIC_REFRESH,
    GENERAL_REFRESH
)
logger.addHandler(logstream)
logger.propagate = False

def moveOldFiles(rootdir, archive, files):
    """
    Move the written files from the root directory given in the config file to the archive

    If keyword DELETE given instead of archive then the flight records will be deleted after being pushed
    """

    # Move the written files from rootdir to the archive
    if archive != "DELETE":
        for file in files:
            path = os.path.join(rootdir, file.split("/")[-1])
            new_path = os.path.join(archive, file.split("/")[-1])
            os.system("mv {} {}".format(path, new_path))
    else:
        for file in files:
            path = os.path.join(rootdir, file.split("/")[-1])
            os.system("rm {}".format(path))

def resolve_link(path):
    logger.debug("Debug: Resolving link for path %s", path)
    mpath = str(path)

    uuid = None
    while len(path.split('/')) > 3 and not uuid:
        pattern = f'http://api.catalogue.ceda.ac.uk/api/v2/observations.json/?fields=uuid,result_field&result_field__dataPath={path}'
        try:
            resp = requests.get(pattern).text
            r = json.loads(resp)
            if r['results']:
                uuid = r['results'][0]['uuid']
                logger.debug("Debug: Reslolving link, found uuid %s", str(uuid))
        except:
            print(f'Unsuccessful link retrieval for {path} - proceeding without')
        path = '/'.join(path.split('/')[:-1])

    if not uuid:
        logger.error("Error: Recursive path search failed for %s", mpath)
        print(f'Recursive path search failed for: {mpath}')

    return uuid

class ESFlightClient(SimpleClient):
    """
    Connects to an elasticsearch instance and exports the
    documents to elasticsearch.
    """

    def __init__(self, index, es_config: str, stac_template: str = None):
        logger.info("Initialising ES Flight Client")

        stac_template = stac_template or STAC_TEMPLATE

        super().__init__(index, es_config=es_config)

        if isinstance(stac_template, str):
            with open(stac_template) as f:
                logger.info(" > Reading stac templace JSON file")
                self.required_keys = json.load(f).keys()

    def push_flights(self, file_list):
        flight_list = []
        if isinstance(file_list[0], str):
            for file in file_list:
                with open(file) as f:
                    flight_list.append(json.load(f))

        elif isinstance(file_list[0], dict):
            flight_list = file_list
        else:
            logger.error("Error: Flight file not found %s", str(file_list[0]))
            raise FileNotFoundError(file_list[0])
        logger.info(f"{len(flight_list)} Flights to be pushed")
        self.push_records(flight_list)
        
    def preprocess_records(self, file_list):
        logger.debug("Debug: Processing following records - %s", file_list)
        
        def set_defaults(refs):
            collection = refs['collection']
            flight_num = refs['properties']['flight_num']
            pcode = refs['properties']['pcode'][0]
            date = refs['properties']['pcode'][1]

            id = f'{collection}__{flight_num}__{pcode}__{date}'.replace(' ','')
            refs['id'] = id
            refs['type'] = 'Feature'
            refs['stac_version'] = '1.0.0'
            refs['stac_extensions'] = [""]
            refs['assets'] = {}
            refs['links'] = []
            return refs

        self.linked = 0
        self.total = len(file_list)
        records = []
        for fid, refs in enumerate(file_list):
            if '_source' in refs.keys():
                id = refs['_id']
                source = refs['_source']
                source['es_id'] = id
            else:
                source = refs

            source = set_defaults(source)

            if 'es_id' in source.keys():
                if source['es_id']:
                    id = source['es_id']
                else:
                    id = gen_id()
                    source['es_id'] = id
            else:
                id = gen_id()
                source['es_id'] = id

            source['catalogue_link'] = ''
            link = resolve_link(source['description_path'])
            if link:
                source['catalogue_link'] = f'https://catalogue.ceda.ac.uk/uuid/{link}'
                self.linked += 1
            else:
                del source['catalogue_link']

            missing = []
            for rq in self.required_keys:
                if rq not in source:
                    missing.append(rq)
            if len(missing) > 0:
                raise TypeError(f"Flight {fid} is missing entries: {missing}")

            source['last_update'] = datetime.strftime(datetime.now(),'%Y-%m-%d %H:%M:%S')

            records.append(source)
        return records

    def obtain_field(self, id, fieldnames):
        logger.info("Info: Performing query to obtain the following fields: %s", str(fieldnames))
        search = {
            "_source": fieldnames,
            "query": {
                "bool":{
                    "must":[{"term":{"id":i}} for i in id]
                }
            }
        }

        resp = self.es.search(
            index=self.index,
            body=search)

        try:
            logger.info("Info: Found following fields: %s", str(resp['hits']['hits'][0]))
            return resp['hits']['hits'][0]
        except IndexError: # No entry found
            logger.error("Error: No entry found.")
            return None

    def add_field(self, id, data, fieldname):
        logger.debug("Debug: Update mapping for id - %s", str(id))
        # Update mapping
        self.es.update(index=self.index, doc_type='_doc', id=id, body={'doc':{fieldname:data}})

    def sort_codes(self):

        ids = [i['_source']['id'] for i in self.pull_records()]

        self.ptcodes = {}
        self.ys, self.yms, self.ymds = {},{},{}

        for ptcode in ids:

            try:
                yr = ptcode.split('*')[1].split('-')[0]
                mth = ptcode.split('*')[1].split('-')[1]
                day = ptcode.split('*')[1].split('-')[2]
            except IndexError:
                try:
                    delim = '__'
                    date_index = 3
                    yr = ptcode.split(delim)[date_index].split('-')[0]
                    mth = ptcode.split(delim)[date_index].split('-')[1]
                    day = ptcode.split(delim)[date_index].split('-')[2]
                except IndexError:
                    pass
            self.ptcodes[ptcode] = 1
            self.ys[yr] = 1
            self.yms[yr + '-' + mth] = 1
            self.ymds[yr + '-' + mth + '-' + day] = 1

    def check_set(self, paths):
        for x, p in enumerate(paths):
            try:
                pcode = re.search(f'[a-z]\d\d\d', p).group()# Replace with re.match at some point
            except AttributeError:
                continue
            search = {
                #"_source": fieldnames,
                "query": {
                    "bool":{
                        "filter":{
                            "bool":{
                                "must":{"term":{"properties.flight_num":pcode}}
                            }
                        }
                    }
                }
            }
            resp = self.es.search(
                index=self.index,
                body=search)

            if resp['hits']['total']['value'] > 0:
                if 'coordinates' in resp['hits']['hits'][0]['geometry']['display']:
                    print(f'{x+1}. {pcode} appears to have coordinates')
                else:
                    print(f'{x+1}. {pcode} appears to not have coordinates')
            else:
                print(f'{x+1}. {pcode} does not have an entry')

    def check_ptcode(self, ptcode):
        if '*' in ptcode:
            return 300
        delim = '__'
        date_index = 3
        print(ptcode)
        yr = ptcode.split(delim)[date_index].split('-')[0]
        mth = ptcode.split(delim)[date_index].split('-')[1]
        day = ptcode.split(delim)[date_index].split('-')[2]
        if yr not in self.ys:
            return 200
        if yr + '-' + mth not in self.yms:
            return 200
        if yr + '-' + mth + '-' + day not in self.ymds:
            return 200
        if ptcode not in self.ptcodes:
            return 200

        # If all filters have passed
        return 100

    def reindex(self, new_index):
        logger.debug("Debug: Reindex for source %s and destination %s", self.index, new_index)

        self.es.reindex({
            "source":{
                "index":self.index
            },
            "dest":{
                "index": new_index
            }#, # Need to carry everything over except geometries.
            #"script":{
            #    "inline":"ctx._source.data.properties.geometry.remove('geometries')"
            #}
        })

    def addFlights(self, archive, rootdir, repush=False, move_files=False):
        """Function for adding flights"""
    
        checked_list = []

        if repush:
            files_list = glob.glob(f'{archive}/*')
        else:
            files_list = glob.glob(f'{rootdir}/*')

        # All flights ok to repush - handled by new client.
        checked_list = list(files_list)

        # Push new flights to index
        logger.info("Identified {} flights".format(len(checked_list)))
        if len(checked_list) > 0:
            self.push_flights(checked_list)
            logger.info("Pushed flights to ES Index")
            if not repush and move_files:
                moveOldFiles(rootdir, archive, checked_list)
                logger.info("Removed local files from push directory - moved to archive")
        else:
            logger.info("Exiting flight pipeline")

    def updateFlights(self, update):
        """
        Update flights using resolve_link() from flightpipe/flight_client module.
        """
        from ceda_flight_pipeline import updaters
        updaters[update](self)

    def refresh(self, refresh_level: int):
        """
        Refresh the catalogue links for specific or all records.
        """

        update_recs = []
        query = GENERAL_REFRESH
        if refresh_level == 1:
            query = SPECIFIC_REFRESH
            logger.info(' > Using specific query for refresh')

        logger.info('Performing index refresh')
        records = self.es.search(index=self.index, query=query, size=10000)

        logger.info(f'Search returned {len(records["hits"]["hits"])} records')

        for record in records['hits']['hits']:
            logger.info(f'Checking {record["_source"]["description_path"]}')
            source = record['_source']
            link = resolve_link(source['description_path'])

            if link:
                logger.info(' > Found valid MOLES link')

                if 'catalogue_link' not in source:
                    logger.info(' >> Confirmed as new link')
                    source['catalogue_link'] = f'https://catalogue.ceda.ac.uk/uuid/{link}'
                    update_recs.append(source)
                    continue

                if link != source['catalogue_link']:
                    logger.info(' >> Confirmed as updated link')
                    source['catalogue_link'] = f'https://catalogue.ceda.ac.uk/uuid/{link}'
                    update_recs.append(source)
                    continue

        if len(update_recs) > 0:
            logger.info(f'Updating {len(update_recs)} records')
            for rec in update_recs:
                self.es.update(
                    index=self.index,
                    id=rec['es_id'],
                    body={'doc':rec,'doc_as_upsert':True}
                )

if __name__ == "__main__":
    print(__file__)



'''
 --- Get Archive Metadata ---
  - Extract data from archive files
  - Copied from elastic_scrapers/scraper_code
'''

import logging

logger = logging.getLogger(__name__)
from ceda_flight_pipeline.utils import logstream
logger.addHandler(logstream)
logger.propagate = False

from ceda_flight_pipeline.utils import jsonWrite

def addList(item,id):
    buffer = []
    try:
        if type(item[id]) != list:
            buffer = [item[id]]
        else:
            buffer = item[id]
    except:
        pass
    return buffer

class FlightConverter:
    def __init__(self, rootdir):
        
        self.rootdir = rootdir
        self.collection_dict = {}
        self.outdir = 'flight_catalog'

        self.scores = {}

    def convert(self):

        self.findFilesRecursive(self.rootdir, 0)

        #self.createCollections()
        #self.createCatalog()

        for score in sorted(self.scores.keys()):
            print(score, self.scores[score])

    def findFilesRecursive(self, directory, r):
        files = os.listdir(directory)
        print('Found {} items'.format(len(files)))
        for idx, f in enumerate(files):
            if r > 0:
                print(' -- ',end='')
            print(idx+1, len(files), f)
            if os.path.isdir(directory + '/' + f):
                self.findFilesRecursive(directory + '/' + f,1)
            else:
                if f.endswith('.json'):
                    filename = directory + '/' + f
                    fileout = f
                    self.convertFile(filename, fileout)

    def convertFile(self, filename, fileout):
        f = open(filename,'r')
        content = json.load(f)
        f.close()

        score = 0
        flight = content["_source"]

        id = filename.split('_flight_info')[0].split('/')[-1]

        try:
            org = flight["misc"]["organisation"]
        except:
            org = flight["file"]["path"].split('/')[2]

        if type(org) == list:
            org = '-'.join(org)

        try:
            pi = flight["misc"]["crew"]["principle"]
        except:
            try:
                pi = flight["misc"]["pi"]
            except:
                pi = ""
        
        if pi != '':
            score += 1

        try:
            airc = flight["misc"]["aircraft"]
        except:
            airc = ""
        
        if airc != "":
                score += 1

        variables = addList(flight["misc"],"variables")

        if variables != "" and variables != []:
            score += 1
        
        location = addList(flight["misc"],"location")

        if location != "" and location != []:
            score += 1

        try:
            platform = flight["misc"]["platform"]
        except:
            platform = ""

        if platform != "":
            score += 1

        instruments = addList(flight["misc"],"instruments")

        if instruments != "" and instruments != []:
            score += 1

        try:
            pcode = flight["misc"]["pcode"]
        except:
            pcode = id

        try:
            alt = flight["misc"]["altitude"]
        except:
            alt = ""

        if alt != "":
            score += 1

        stac_content = {
            "id": id,
            "type" : "Feature",
            "stac_version": "1.0.0",
            "stac_extensions":[""],

            "es_id":content["_id"],

            "description_path":flight["file"]["path"],

            "collection":org,

            "geometry": flight["spatial"]["geometries"],
            "properties":{
                "data_format":flight["data_format"]["format"],
                "start_datetime":flight["temporal"]["start_time"],
                "end_datetime": flight["temporal"]["end_time"],
                "flight_num":flight["misc"]["flight_num"],
                "pcode": pcode,
                "aircraft": airc,
                "altitude":alt,
                "variables": variables,
                "location": location,
                "platform": platform,
                "instruments": instruments,
                "principle": pi
            },
            "assets":{},
            "links":[]
        }

        coords = flight["spatial"]["geometries"]["display"]["coordinates"]
        non_emp = -1
        try:
            is_ne = False
            while not is_ne:
                non_emp += 1
                if len(coords[non_emp]) != 0:
                    is_ne = True

            if len(coords[non_emp]) == 2: # Line String
                coords_lat = np.array(coords)[:,0]
                coords_lon = np.array(coords)[:,1]
            else:
                stac_content["geometry"]["display"]["type"] = "MultiLineString"

                coords_lat = np.array(coords[non_emp])[:,0]
                coords_lon = np.array(coords[non_emp])[:,1]

                for x in range(non_emp+1,len(coords)):
                    cline = coords[x]
                    if len(cline) > 0:
                        cline_lat = np.array(cline)[:,0]
                        cline_lon = np.array(cline)[:,1]

                        coords_lat = np.concatenate((coords_lat, cline_lat))
                        coords_lon = np.concatenate((coords_lon, cline_lon))

        except KeyError:
            print('Issue')
            x=input()
        latmax = np.max(coords_lat)
        latmin = np.min(coords_lat)
        lonmax = np.max(coords_lon)
        lonmin = np.min(coords_lon)


        try:
            orgmeta = self.collection_dict[org]
            orgmeta['items'].append(id)

            if latmin < orgmeta['spatial_minmax'][0]:
                orgmeta['spatial_minmax'][0] = latmin
            if latmax > orgmeta['spatial_minmax'][1]:
                orgmeta['spatial_minmax'][1] = latmax
            if lonmin < orgmeta['spatial_minmax'][2]:
                orgmeta['spatial_minmax'][2] = lonmin
            if lonmax > orgmeta['spatial_minmax'][3]:
                orgmeta['spatial_minmax'][3] = lonmax

            tmin = [flight["temporal"]["start_time"], orgmeta['temporal_minmax'][0]]
            tmax = [flight["temporal"]["end_time"], orgmeta['temporal_minmax'][1]]

            orgmeta['temporal_minmax'] = [
                sorted(tmin)[0],
                sorted(tmax)[1]
            ]

            self.collection_dict[org] = orgmeta

        except KeyError:
            self.collection_dict[org] = {
                'items':[id],
                'spatial_minmax':[latmin, latmax, lonmin, lonmax], # min_lat ... max_lon
                'temporal_minmax':[
                    flight["temporal"]["start_time"],
                    flight["temporal"]["end_time"]
                    ]
            }

        try:
            self.scores[str(score)] += 1
        except:
            self.scores[str(score)] = 1

        jsonWrite('{}/collections/items/'.format(self.outdir), fileout, stac_content)

    def createCollections(self):

        for org in self.collection_dict.keys():

            orgmeta = self.collection_dict[org]

            desc = '{} Flight Collection'.format(org)
            cfile = "collections/{}.json".format(org)

            ## Build Links

            collection_links = [{
                    "rel":"self",
                    "href":cfile,
                    "type":"application/json",
                    "title":org
                }
            ]
            for id in orgmeta['items']:
                collection_links.append({
                    "rel":"item",
                    "href":"items/{}_flight_info.json".format(id), # Filename
                    "type":"application/json",
                    "title":id
                })

            collection = {
                "id":org,
                "type": "Collection",
                "stac_extensions": [
                    ""
                ],
                "stac_version":"1.0.0",
                "description": desc,
                "title": org,
                "providers": [],
                "extent": {
                    "spatial": {"bbox":[orgmeta['spatial_minmax']]},
                    "temporal":{"interval":[orgmeta['temporal_minmax']]}
                },
                "license":"", # Find this
                "summaries":{},
                "links":collection_links
            }

            jsonWrite(self.outdir,cfile, collection)

    def createCatalog(self):

        catalog_links = [
            {
                "rel":"self",
                "href":"./flight_catalog.json",
                "type":"application/json",
                "title":"Flight Catalog"
            }
        ]

        for org in self.collection_dict.keys():
            catalog_links.append({
                "rel":"child",
                "href":"collections/{}.json".format(org), # Filename
                "type":"application/json",
                "title":org
            })

        catalog = {
            "id":"Flight_Catalog",
            "type":"Catalog",
            "stac_version":"1.6.1",
            "stac_extensions":[""],
            "title":"Flight Catalog",
            "description": "Full catalog of flights for CEDA flight finder",
            "links":catalog_links
        }

        jsonWrite(self.outdir,"flight_catalog", catalog)




import os, sys
import json
import numpy as np

import logging
logger = logging.getLogger(__name__)

logging.basicConfig(level=logging.WARNING)
logstream = logging.StreamHandler()

formatter = logging.Formatter('%(levelname)s [%(name)s]: %(message)s')
logstream.setFormatter(formatter)

STAC_TEMPLATE = {
    "id": None,
    "es_id": None,
    "type" : "Feature",
    "stac_version": "1.0.0",
    "stac_extensions":[""],
    "description_path":None,
    "collection":None,
    "geometry": None,
    "properties":None,
    "assets":{},
    "links":[]
}

SPECIFIC_REFRESH = {
    "bool": {
        "must_not": [
            {
                "exists": {
                    "field": "catalogue_link"
                }
            }
        ]
    }
}

GENERAL_REFRESH = {
    "match_all":{},
}

def jsonWrite(path, file, content):
    if not os.path.exists(path):
        os.makedirs(path)

    g = open(path + '/' + file,'w')
    g.write(json.dumps(content))
    g.close()

def jsonRead(path, file):
    g = open(path + '/' + file, 'r')
    content = json.load(g)
    g.close()
    return content

def recursiveList(obj, lv):
    for element in obj:
        for x in range(lv):
            print(' - ',end='')
        if type(element) == list:
            print('list')
            recursiveList(element, lv+1)
        else:
            print(type(element))

def recursiveConvert(obj):
    # Returns a complete numpy array from a ragged list
    out_arr = None
    is_init = False
    for element in obj:
        if element != []:
            if type(element) == list and len(element) != 2:
                coord = recursiveConvert(element)
            else:
                coord = np.array(element)

            if not is_init:
                out_arr = coord
                is_init = True
            else:
                out_arr = np.vstack((out_arr, coord))
    return out_arr

def forceSearch(obj, maxs, mins, depth):
    perform_checks = False
    if len(obj) == 2 and type(obj) == list:
        if type(obj[0]) != list:
            perform_checks = True

    if perform_checks:
        if obj[0] > maxs[0]:
            maxs[0] = obj[0]
        if obj[0] < mins[0]:
            mins[0] = obj[0]

        if obj[1] > maxs[1]:
            maxs[1] = obj[1]
        if obj[1] < mins[1]:
            mins[1] = obj[1]
        return [maxs, mins]
    else:
        for ob in obj:
            [maxs, mins] = forceSearch(ob, maxs, mins, depth+1)
    return [maxs, mins]
            
def genID():
    import random
    chars = [*'0123456789abcdefghijklmnopqrstuvwxyz']
    id = ''
    for i in range(39):
        j = random.randint(0,len(chars)-1)
        id += chars[j]

    os.system(f'grep {id} python_scripts/cache/id_history > python_scripts/cache/matches')
    with open('python_scripts/cache/matches') as f:
        content = f.readlines()
        if len(content) > 0:
            print("""
CONGRATULATIONS!!! If you are reading this message you have just computed a flight index number
that matches 39 out of the 40 characters of at least one other flight index number, the probability of which was
                  1 in 4.9 x 10^60
Meaning that unless you've broken the randomiser function you're extremely lucky today and should immediately go buy a lottery ticket!
                  
So lucky in fact that if you'd been running this program non-stop since the beginning of the universe on 10 billion computers, 
generating a flight ID every nano second, the probability of what just occurred is about as likely as you winning the EuroMillions 
                  lottery every day for 5 days
                  """)
            raise ValueError('Probabilty Factor Exceeds human tolerances, please rerun program')
        else:
            id += chars[random.randint(0, len(chars))]
    with open('python_scripts/cache/id_history', 'a') as f:
        f.write(id + '\n')
    print(id)
    return id


import unittest


class TestImports(unittest.TestCase):
    def test_import_math(self):
        try:
            import click
        except ImportError:
            self.fail("Failed to import math module")

    def test_import_os(self):
        try:
            import os
        except ImportError:
            self.fail("Failed to import os module")

    def test_import_numpy(self):
        try:
            import numpy
        except ImportError:
            self.fail("Failed to import numpy module")

    def test_import_poetry(self):
        try:
            import poetry
        except ImportError:
            self.fail("Failed to import poetry module")

    def test_import_elasticsearch(self):
        try:
            import elasticsearch
        except ImportError:
            self.fail("Failed to import elasticsearch module")

    def test_import_urllib3(self):
        try:
            import urllib3
        except ImportError:
            self.fail("Failed to import urllib3 module")

    def test_import_certifi(self):
        try:
            import certifi
        except ImportError:
            self.fail("Failed to import certifi module")


if __name__ == "__main__":
    unittest.main()



import json
import os
from typing import Union

from elasticsearch import Elasticsearch
from elasticsearch.helpers import bulk

import logging
logger = logging.getLogger(__name__)
from ceda_flight_pipeline.utils import logstream
logger.addHandler(logstream)
logger.propagate = False

def gen_id():
    import random
    chars = [*'0123456789abcdefghijklmnopqrstuvwxyz']
    xid = ''
    for i in range(39):
        j = random.randint(0,len(chars)-1)
        xid += chars[j]
    return xid
    # Probability of id reuse is negligible.

default_rec = {
    "_index": None,
    "_type": "_doc",
    "_id":None,
    "_score": 1.0,
    "_source":None
}

settings_default = {
    "hosts": [
        "https://elasticsearch.ceda.ac.uk"
    ],
    "headers": {
        "x-api-key": ""
    }
}

def create_settings(es_config='es_settings.json'):
    with open(es_config,'w') as f:
        f.write(json.dumps(settings_default))
    

class ResponseException(Exception):
    """
    Response Exception - Elasticsearch did not return an object with any 'hits'. This indicates an error,
    as even with 0 hits returned there should still be an empty 'hits' list here.
    """
    def __init__(self, keys=None):
        self.message = f'Received ({keys}) keys from response - No "hits" received'
        super().__init__(self.message)

class SimpleClient:
    """
    Simple Elasticsearch-Python client for bulk operations
    """
    def __init__(self, index: str, es_config: Union[dict,str]):
        """
        Initialise client, pull credentials from a configuration file if present
        and create an underlying client within this class.
        """
        self.index = index

        if isinstance(es_config,str):
            if not os.path.isfile(es_config):
                raise FileNotFoundError(f'File {es_config} not present, no settings loaded.')
            with open(es_config) as f:
                connection_kwargs = json.load(f)
        else:
            connection_kwargs = es_config

        self.es = Elasticsearch(**connection_kwargs)

    def preprocess_records(self, records):
        """
        Blank preprocessor method - override with preprocessing steps for specific
        client use-cases.
        """
        return records

    def process_records(self, records):
        """
        Final processing step to add ids to new records if they are not already present.
        """
        for r in records:
            yield r

    def get_size(self):
        md = self.es.cat.count(index=self.index, params={"format": "json"})[0]
        return md['count']

    def push_records(self, records):
        """
        Push records using bulk helper tool.
        """
        records = self.preprocess_records(records)
        for r in records:
            self.es.update(
                index=self.index,
                id=r['es_id'],
                body={'doc':r,'doc_as_upsert':True}
            )

    def pull_records(self):
        """
        Pull all records up to 10,000 and return all hits.
        """
        search = {
            "size":10000,
            "query": {
                "match_all":{}
            }
        }

        resp = self.es.search(
            index=self.index,
            body=search)

        try:
            return resp['hits']['hits']
        except IndexError:
            raise ResponseException(keys=resp.keys())

        